@@ -12,17 +12,23 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-from __future__ import absolute_import, division, print_function
-
 import collections
+import datetime
+import types
 
 from tornado import gen, ioloop
-from tornado.concurrent import Future
+from tornado.concurrent import Future, future_set_result_unless_cancelled
 
-__all__ = ['Condition', 'Event', 'Semaphore', 'BoundedSemaphore', 'Lock']
+from typing import Union, Optional, Type, Any, Awaitable
+import typing
+
+if typing.TYPE_CHECKING:
+    from typing import Deque, Set  # noqa: F401
+
+__all__ = ["Condition", "Event", "Semaphore", "BoundedSemaphore", "Lock"]
 
 
-class _TimeoutGarbageCollector(object):
+class _TimeoutGarbageCollector:
     """Base class for objects that periodically clean up timed-out waiters.
 
     Avoids memory leak in a common pattern like:
@@ -31,17 +37,17 @@ class _TimeoutGarbageCollector(object):
             yield condition.wait(short_timeout)
             print('looping....')
     """
-    def __init__(self):
-        self._waiters = collections.deque()  # Futures.
+
+    def __init__(self) -> None:
+        self._waiters = collections.deque()  # type: Deque[Future]
         self._timeouts = 0
 
-    def _garbage_collect(self):
+    def _garbage_collect(self) -> None:
         # Occasionally clear timed-out waiters.
         self._timeouts += 1
         if self._timeouts > 100:
             self._timeouts = 0
-            self._waiters = collections.deque(
-                w for w in self._waiters if not w.done())
+            self._waiters = collections.deque(w for w in self._waiters if not w.done())
 
 
 class Condition(_TimeoutGarbageCollector):
@@ -54,30 +60,27 @@ class Condition(_TimeoutGarbageCollector):
 
     .. testcode::
 
+        import asyncio
         from tornado import gen
-        from tornado.ioloop import IOLoop
         from tornado.locks import Condition
 
         condition = Condition()
 
-        @gen.coroutine
-        def waiter():
+        async def waiter():
             print("I'll wait right here")
-            yield condition.wait()  # Yield a Future.
+            await condition.wait()
             print("I'm done waiting")
 
-        @gen.coroutine
-        def notifier():
+        async def notifier():
             print("About to notify")
             condition.notify()
             print("Done notifying")
 
-        @gen.coroutine
-        def runner():
-            # Yield two Futures; wait for waiter() and notifier() to finish.
-            yield [waiter(), notifier()]
+        async def runner():
+            # Wait for waiter() and notifier() in parallel
+            await gen.multi([waiter(), notifier()])
 
-        IOLoop.current().run_sync(runner)
+        asyncio.run(runner())
 
     .. testoutput::
 
@@ -92,46 +95,50 @@ class Condition(_TimeoutGarbageCollector):
         io_loop = IOLoop.current()
 
         # Wait up to 1 second for a notification.
-        yield condition.wait(timeout=io_loop.time() + 1)
+        await condition.wait(timeout=io_loop.time() + 1)
 
     ...or a `datetime.timedelta` for a timeout relative to the current time::
 
         # Wait up to 1 second.
-        yield condition.wait(timeout=datetime.timedelta(seconds=1))
+        await condition.wait(timeout=datetime.timedelta(seconds=1))
 
-    The method raises `tornado.gen.TimeoutError` if there's no notification
-    before the deadline.
+    The method returns False if there's no notification before the deadline.
+
+    .. versionchanged:: 5.0
+       Previously, waiters could be notified synchronously from within
+       `notify`. Now, the notification will always be received on the
+       next iteration of the `.IOLoop`.
     """
 
-    def __init__(self):
-        super(Condition, self).__init__()
-        self.io_loop = ioloop.IOLoop.current()
-
-    def __repr__(self):
-        result = '<%s' % (self.__class__.__name__, )
+    def __repr__(self) -> str:
+        result = f"<{self.__class__.__name__}"
         if self._waiters:
-            result += ' waiters[%s]' % len(self._waiters)
-        return result + '>'
+            result += " waiters[%s]" % len(self._waiters)
+        return result + ">"
 
-    def wait(self, timeout=None):
+    def wait(
+        self, timeout: Optional[Union[float, datetime.timedelta]] = None
+    ) -> Awaitable[bool]:
         """Wait for `.notify`.
 
         Returns a `.Future` that resolves ``True`` if the condition is notified,
         or ``False`` after a timeout.
         """
-        waiter = Future()
+        waiter = Future()  # type: Future[bool]
         self._waiters.append(waiter)
         if timeout:
-            def on_timeout():
-                waiter.set_result(False)
+
+            def on_timeout() -> None:
+                if not waiter.done():
+                    future_set_result_unless_cancelled(waiter, False)
                 self._garbage_collect()
+
             io_loop = ioloop.IOLoop.current()
             timeout_handle = io_loop.add_timeout(timeout, on_timeout)
-            waiter.add_done_callback(
-                lambda _: io_loop.remove_timeout(timeout_handle))
+            waiter.add_done_callback(lambda _: io_loop.remove_timeout(timeout_handle))
         return waiter
 
-    def notify(self, n=1):
+    def notify(self, n: int = 1) -> None:
         """Wake ``n`` waiters."""
         waiters = []  # Waiters we plan to run right now.
         while n and self._waiters:
@@ -141,14 +148,14 @@ class Condition(_TimeoutGarbageCollector):
                 waiters.append(waiter)
 
         for waiter in waiters:
-            waiter.set_result(True)
+            future_set_result_unless_cancelled(waiter, True)
 
-    def notify_all(self):
+    def notify_all(self) -> None:
         """Wake all waiters."""
         self.notify(len(self._waiters))
 
 
-class Event(object):
+class Event:
     """An event blocks coroutines until its internal flag is set to True.
 
     Similar to `threading.Event`.
@@ -158,30 +165,27 @@ class Event(object):
 
     .. testcode::
 
+        import asyncio
         from tornado import gen
-        from tornado.ioloop import IOLoop
         from tornado.locks import Event
 
         event = Event()
 
-        @gen.coroutine
-        def waiter():
+        async def waiter():
             print("Waiting for event")
-            yield event.wait()
+            await event.wait()
             print("Not waiting this time")
-            yield event.wait()
+            await event.wait()
             print("Done")
 
-        @gen.coroutine
-        def setter():
+        async def setter():
             print("About to set the event")
             event.set()
 
-        @gen.coroutine
-        def runner():
-            yield [waiter(), setter()]
+        async def runner():
+            await gen.multi([waiter(), setter()])
 
-        IOLoop.current().run_sync(runner)
+        asyncio.run(runner())
 
     .. testoutput::
 
@@ -190,60 +194,88 @@ class Event(object):
         Not waiting this time
         Done
     """
-    def __init__(self):
-        self._future = Future()
 
-    def __repr__(self):
-        return '<%s %s>' % (
-            self.__class__.__name__, 'set' if self.is_set() else 'clear')
+    def __init__(self) -> None:
+        self._value = False
+        self._waiters = set()  # type: Set[Future[None]]
 
-    def is_set(self):
+    def __repr__(self) -> str:
+        return "<{} {}>".format(
+            self.__class__.__name__,
+            "set" if self.is_set() else "clear",
+        )
+
+    def is_set(self) -> bool:
         """Return ``True`` if the internal flag is true."""
-        return self._future.done()
+        return self._value
 
-    def set(self):
+    def set(self) -> None:
         """Set the internal flag to ``True``. All waiters are awakened.
 
         Calling `.wait` once the flag is set will not block.
         """
-        if not self._future.done():
-            self._future.set_result(None)
+        if not self._value:
+            self._value = True
 
-    def clear(self):
+            for fut in self._waiters:
+                if not fut.done():
+                    fut.set_result(None)
+
+    def clear(self) -> None:
         """Reset the internal flag to ``False``.
 
         Calls to `.wait` will block until `.set` is called.
         """
-        if self._future.done():
-            self._future = Future()
+        self._value = False
 
-    def wait(self, timeout=None):
+    def wait(
+        self, timeout: Optional[Union[float, datetime.timedelta]] = None
+    ) -> Awaitable[None]:
         """Block until the internal flag is true.
 
-        Returns a Future, which raises `tornado.gen.TimeoutError` after a
+        Returns an awaitable, which raises `tornado.util.TimeoutError` after a
         timeout.
         """
+        fut = Future()  # type: Future[None]
+        if self._value:
+            fut.set_result(None)
+            return fut
+        self._waiters.add(fut)
+        fut.add_done_callback(lambda fut: self._waiters.remove(fut))
         if timeout is None:
-            return self._future
+            return fut
         else:
-            return gen.with_timeout(timeout, self._future)
+            timeout_fut = gen.with_timeout(timeout, fut)
+            # This is a slightly clumsy workaround for the fact that
+            # gen.with_timeout doesn't cancel its futures. Cancelling
+            # fut will remove it from the waiters list.
+            timeout_fut.add_done_callback(
+                lambda tf: fut.cancel() if not fut.done() else None
+            )
+            return timeout_fut
 
 
-class _ReleasingContextManager(object):
+class _ReleasingContextManager:
     """Releases a Lock or Semaphore at the end of a "with" statement.
 
-        with (yield semaphore.acquire()):
-            pass
-
-        # Now semaphore.release() has been called.
-    """
-    def __init__(self, obj):
-        self._obj = obj
-
-    def __enter__(self):
+    with (yield semaphore.acquire()):
         pass
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    # Now semaphore.release() has been called.
+    """
+
+    def __init__(self, obj: Any) -> None:
+        self._obj = obj
+
+    def __enter__(self) -> None:
+        pass
+
+    def __exit__(
+        self,
+        exc_type: "Optional[Type[BaseException]]",
+        exc_val: Optional[BaseException],
+        exc_tb: Optional[types.TracebackType],
+    ) -> None:
         self._obj.release()
 
 
@@ -266,44 +298,48 @@ class Semaphore(_TimeoutGarbageCollector):
        from tornado.ioloop import IOLoop
        from tornado.concurrent import Future
 
-       # Ensure reliable doctest output: resolve Futures one at a time.
-       futures_q = deque([Future() for _ in range(3)])
+       inited = False
 
-       @gen.coroutine
-       def simulator(futures):
+       async def simulator(futures):
            for f in futures:
-               yield gen.moment
+               # simulate the asynchronous passage of time
+               await gen.sleep(0)
+               await gen.sleep(0)
                f.set_result(None)
 
-       IOLoop.current().add_callback(simulator, list(futures_q))
-
        def use_some_resource():
+           global inited
+           global futures_q
+           if not inited:
+               inited = True
+               # Ensure reliable doctest output: resolve Futures one at a time.
+               futures_q = deque([Future() for _ in range(3)])
+               IOLoop.current().add_callback(simulator, list(futures_q))
+
            return futures_q.popleft()
 
     .. testcode:: semaphore
 
+        import asyncio
         from tornado import gen
-        from tornado.ioloop import IOLoop
         from tornado.locks import Semaphore
 
         sem = Semaphore(2)
 
-        @gen.coroutine
-        def worker(worker_id):
-            yield sem.acquire()
+        async def worker(worker_id):
+            await sem.acquire()
             try:
                 print("Worker %d is working" % worker_id)
-                yield use_some_resource()
+                await use_some_resource()
             finally:
                 print("Worker %d is done" % worker_id)
                 sem.release()
 
-        @gen.coroutine
-        def runner():
+        async def runner():
             # Join all workers.
-            yield [worker(i) for i in range(3)]
+            await gen.multi([worker(i) for i in range(3)])
 
-        IOLoop.current().run_sync(runner)
+        asyncio.run(runner())
 
     .. testoutput:: semaphore
 
@@ -317,7 +353,18 @@ class Semaphore(_TimeoutGarbageCollector):
     Workers 0 and 1 are allowed to run concurrently, but worker 2 waits until
     the semaphore has been released once, by worker 0.
 
-    `.acquire` is a context manager, so ``worker`` could be written as::
+    The semaphore can be used as an async context manager::
+
+        async def worker(worker_id):
+            async with sem:
+                print("Worker %d is working" % worker_id)
+                await use_some_resource()
+
+            # Now the semaphore has been released.
+            print("Worker %d is done" % worker_id)
+
+    For compatibility with older versions of Python, `.acquire` is a
+    context manager, so ``worker`` could also be written as::
 
         @gen.coroutine
         def worker(worker_id):
@@ -328,36 +375,26 @@ class Semaphore(_TimeoutGarbageCollector):
             # Now the semaphore has been released.
             print("Worker %d is done" % worker_id)
 
-    In Python 3.5, the semaphore itself can be used as an async context
-    manager::
-
-        async def worker(worker_id):
-            async with sem:
-                print("Worker %d is working" % worker_id)
-                await use_some_resource()
-
-            # Now the semaphore has been released.
-            print("Worker %d is done" % worker_id)
-
     .. versionchanged:: 4.3
        Added ``async with`` support in Python 3.5.
+
     """
-    def __init__(self, value=1):
-        super(Semaphore, self).__init__()
+
+    def __init__(self, value: int = 1) -> None:
+        super().__init__()
         if value < 0:
-            raise ValueError('semaphore initial value must be >= 0')
+            raise ValueError("semaphore initial value must be >= 0")
 
         self._value = value
 
-    def __repr__(self):
-        res = super(Semaphore, self).__repr__()
-        extra = 'locked' if self._value == 0 else 'unlocked,value:{0}'.format(
-            self._value)
+    def __repr__(self) -> str:
+        res = super().__repr__()
+        extra = "locked" if self._value == 0 else f"unlocked,value:{self._value}"
         if self._waiters:
-            extra = '{0},waiters:{1}'.format(extra, len(self._waiters))
-        return '<{0} [{1}]>'.format(res[1:-1], extra)
+            extra = f"{extra},waiters:{len(self._waiters)}"
+        return f"<{res[1:-1]} [{extra}]>"
 
-    def release(self):
+    def release(self) -> None:
         """Increment the counter and wake one waiter."""
         self._value += 1
         while self._waiters:
@@ -374,41 +411,54 @@ class Semaphore(_TimeoutGarbageCollector):
                 waiter.set_result(_ReleasingContextManager(self))
                 break
 
-    def acquire(self, timeout=None):
-        """Decrement the counter. Returns a Future.
+    def acquire(
+        self, timeout: Optional[Union[float, datetime.timedelta]] = None
+    ) -> Awaitable[_ReleasingContextManager]:
+        """Decrement the counter. Returns an awaitable.
 
-        Block if the counter is zero and wait for a `.release`. The Future
+        Block if the counter is zero and wait for a `.release`. The awaitable
         raises `.TimeoutError` after the deadline.
         """
-        waiter = Future()
+        waiter = Future()  # type: Future[_ReleasingContextManager]
         if self._value > 0:
             self._value -= 1
             waiter.set_result(_ReleasingContextManager(self))
         else:
             self._waiters.append(waiter)
             if timeout:
-                def on_timeout():
-                    waiter.set_exception(gen.TimeoutError())
+
+                def on_timeout() -> None:
+                    if not waiter.done():
+                        waiter.set_exception(gen.TimeoutError())
                     self._garbage_collect()
+
                 io_loop = ioloop.IOLoop.current()
                 timeout_handle = io_loop.add_timeout(timeout, on_timeout)
                 waiter.add_done_callback(
-                    lambda _: io_loop.remove_timeout(timeout_handle))
+                    lambda _: io_loop.remove_timeout(timeout_handle)
+                )
         return waiter
 
-    def __enter__(self):
-        raise RuntimeError(
-            "Use Semaphore like 'with (yield semaphore.acquire())', not like"
-            " 'with semaphore'")
+    def __enter__(self) -> None:
+        raise RuntimeError("Use 'async with' instead of 'with' for Semaphore")
 
-    __exit__ = __enter__
+    def __exit__(
+        self,
+        typ: "Optional[Type[BaseException]]",
+        value: Optional[BaseException],
+        traceback: Optional[types.TracebackType],
+    ) -> None:
+        self.__enter__()
 
-    @gen.coroutine
-    def __aenter__(self):
-        yield self.acquire()
+    async def __aenter__(self) -> None:
+        await self.acquire()
 
-    @gen.coroutine
-    def __aexit__(self, typ, value, tb):
+    async def __aexit__(
+        self,
+        typ: "Optional[Type[BaseException]]",
+        value: Optional[BaseException],
+        tb: Optional[types.TracebackType],
+    ) -> None:
         self.release()
 
 
@@ -420,18 +470,19 @@ class BoundedSemaphore(Semaphore):
     resources with limited capacity, so a semaphore released too many times
     is a sign of a bug.
     """
-    def __init__(self, value=1):
-        super(BoundedSemaphore, self).__init__(value=value)
+
+    def __init__(self, value: int = 1) -> None:
+        super().__init__(value=value)
         self._initial_value = value
 
-    def release(self):
+    def release(self) -> None:
         """Increment the counter and wake one waiter."""
         if self._value >= self._initial_value:
             raise ValueError("Semaphore released too many times")
-        super(BoundedSemaphore, self).release()
+        super().release()
 
 
-class Lock(object):
+class Lock:
     """A lock for coroutines.
 
     A Lock begins unlocked, and `acquire` locks it immediately. While it is
@@ -440,26 +491,24 @@ class Lock(object):
 
     Releasing an unlocked lock raises `RuntimeError`.
 
-    `acquire` supports the context manager protocol in all Python versions:
+    A Lock can be used as an async context manager with the ``async
+    with`` statement:
 
-    >>> from tornado import gen, locks
+    >>> from tornado import locks
     >>> lock = locks.Lock()
     >>>
-    >>> @gen.coroutine
-    ... def f():
-    ...    with (yield lock.acquire()):
+    >>> async def f():
+    ...    async with lock:
     ...        # Do something holding the lock.
     ...        pass
     ...
     ...    # Now the lock is released.
 
-    In Python 3.5, `Lock` also supports the async context manager
-    protocol. Note that in this case there is no `acquire`, because
-    ``async with`` includes both the ``yield`` and the ``acquire``
-    (just as it does with `threading.Lock`):
+    For compatibility with older versions of Python, the `.acquire`
+    method asynchronously returns a regular context manager:
 
-    >>> async def f():  # doctest: +SKIP
-    ...    async with lock:
+    >>> async def f2():
+    ...    with (yield lock.acquire()):
     ...        # Do something holding the lock.
     ...        pass
     ...
@@ -469,23 +518,24 @@ class Lock(object):
        Added ``async with`` support in Python 3.5.
 
     """
-    def __init__(self):
+
+    def __init__(self) -> None:
         self._block = BoundedSemaphore(value=1)
 
-    def __repr__(self):
-        return "<%s _block=%s>" % (
-            self.__class__.__name__,
-            self._block)
+    def __repr__(self) -> str:
+        return f"<{self.__class__.__name__} _block={self._block}>"
 
-    def acquire(self, timeout=None):
-        """Attempt to lock. Returns a Future.
+    def acquire(
+        self, timeout: Optional[Union[float, datetime.timedelta]] = None
+    ) -> Awaitable[_ReleasingContextManager]:
+        """Attempt to lock. Returns an awaitable.
 
-        Returns a Future, which raises `tornado.gen.TimeoutError` after a
+        Returns an awaitable, which raises `tornado.util.TimeoutError` after a
         timeout.
         """
         return self._block.acquire(timeout)
 
-    def release(self):
+    def release(self) -> None:
         """Unlock.
 
         The first coroutine in line waiting for `acquire` gets the lock.
@@ -495,18 +545,26 @@ class Lock(object):
         try:
             self._block.release()
         except ValueError:
-            raise RuntimeError('release unlocked lock')
+            raise RuntimeError("release unlocked lock")
 
-    def __enter__(self):
-        raise RuntimeError(
-            "Use Lock like 'with (yield lock)', not like 'with lock'")
+    def __enter__(self) -> None:
+        raise RuntimeError("Use `async with` instead of `with` for Lock")
 
-    __exit__ = __enter__
+    def __exit__(
+        self,
+        typ: "Optional[Type[BaseException]]",
+        value: Optional[BaseException],
+        tb: Optional[types.TracebackType],
+    ) -> None:
+        self.__enter__()
 
-    @gen.coroutine
-    def __aenter__(self):
-        yield self.acquire()
+    async def __aenter__(self) -> None:
+        await self.acquire()
 
-    @gen.coroutine
-    def __aexit__(self, typ, value, tb):
+    async def __aexit__(
+        self,
+        typ: "Optional[Type[BaseException]]",
+        value: Optional[BaseException],
+        tb: Optional[types.TracebackType],
+    ) -> None:
         self.release()
